#include <sys/types.h>
#include <sys/fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <err.h>
#include <string.h>
#include <libgeom.h>
#include <getopt.h>
#include <md5.h>


#define BUF_SIZE (1024)
#define STEP_SIZE (32*1024)
#define STEP_OFFSET (STEP_SIZE-BUF_SIZE)
#define START_OFFSET 512


char *buf;
size_t buf_size;


int main(int argc, char **argv) 
{
	int fd, fdr;
	size_t sector_size;
	unsigned int i;
	size_t offset;
	char *devname;
	char *buf2;

	if (argc < 2) {
		printf("usage: %s devname\n", argv[0]);
		exit(1);
	}
	devname = argv[1];


	buf = malloc(BUF_SIZE);
	if (buf == NULL)
		err(1, "Cannot allocate %d KB", BUF_SIZE / 1024);

	fdr = open("/dev/random", O_RDONLY);
	if (read(fdr, buf, BUF_SIZE) != BUF_SIZE)
		errx(1, "Cannot fetch %d KB of random data", BUF_SIZE / 1024);
	close(fdr);

	fd = g_open(devname, 1);
	if (fd < 0)
		err(1, "Cannot open %s", devname);
	sector_size = g_sectorsize(fd);

	buf2 = malloc(STEP_SIZE);
	if (buf2 == NULL)
		errx(1, "Cannot allocate %u", STEP_SIZE);

	for (i = 0; i < 1000; i++) {
		offset = START_OFFSET+(i*9)*sector_size;

		/* Write a buffer in the middle of a STEP_SIZE-sized block */
		if (pwrite(fd, buf, BUF_SIZE, offset+STEP_OFFSET) != BUF_SIZE)
			err(1, "Cannot read %u at %u", BUF_SIZE, STEP_OFFSET);

		/* Read a whole block */
		if (pread(fd, buf2, STEP_SIZE, offset) != STEP_SIZE)
			err(1, "Cannot read %u at %u", STEP_SIZE, offset);

		/* Compare the written buffer in the middle with the original buffer */
		if (bcmp(buf2+STEP_OFFSET, buf, BUF_SIZE) != 0)
			errx(1, "CMP FAILED at %u", offset);
	}

	/* Replay only the read requests */
	for (i = 0; i < 1000; i++) {
		offset = START_OFFSET+(i*9)*sector_size;

		/* Read a whole block */
		if (pread(fd, buf2, STEP_SIZE, offset) != STEP_SIZE)
			err(1, "Cannot read %u at %u", STEP_SIZE, offset);

		/* Compare the written buffer in the middle with the original buffer */
		if (bcmp(buf2+STEP_OFFSET, buf, BUF_SIZE) != 0)
			errx(1, "CMP FAILED at %u", offset);
	}

	
	return 0;
}
