#include <sys/types.h>
#include <sys/fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <err.h>
#include <string.h>
#include <libgeom.h>
#include <getopt.h>
#include <md5.h>


#define BUF_SIZE (1024*1)
#define BIG_BUF_SIZE (1024*128)
#define START_OFFSET (1024*64)


void gen_random(char *buf, size_t size)
{
	int fdr;
	fdr = open("/dev/random", O_RDONLY);
	if (read(fdr, buf, size) != size)
		errx(1, "Cannot fetch %d KB of random data", size / 1024);
	close(fdr);
}


int main(int argc, char **argv) 
{
	int fd;
	size_t sector_size;
	char *devname;
	char *buf;
	char *buf2;
	size_t offset, big_offset;

	if (argc < 2) {
		printf("usage: %s devname\n", argv[0]);
		exit(1);
	}
	devname = argv[1];


	buf = malloc(BUF_SIZE);
	if (buf == NULL)
		err(1, "Cannot allocate %d KB", BUF_SIZE / 1024);

	buf2 = malloc(BIG_BUF_SIZE);
	if (buf2 == NULL)
		err(1, "Cannot allocate %d KB", BIG_BUF_SIZE / 1024);

	fd = g_open(devname, 1);
	if (fd < 0)
		err(1, "Cannot open %s", devname);
	sector_size = g_sectorsize(fd);

	offset = big_offset = 0;
	while (offset+BUF_SIZE <= big_offset+BIG_BUF_SIZE) {
		gen_random(buf, BUF_SIZE);
		/* write a small buffer */
		if (pwrite(fd, buf, BUF_SIZE, offset+START_OFFSET) != BUF_SIZE)
			err(1, "Cannot write %u at %u", BUF_SIZE, offset+START_OFFSET);
		/* read a big buffer containing the small buffer */
		if (pread(fd, buf2, BIG_BUF_SIZE, big_offset+START_OFFSET) != BIG_BUF_SIZE)
			err(1, "Cannot read %u at %u", BIG_BUF_SIZE, big_offset+START_OFFSET);
		/* compare */
		if (bcmp(buf, buf2+offset, BUF_SIZE) != 0)
			errx(1, "COMPARE FAILED AT %u", offset);
		offset += sector_size;
	}
	
	return 0;
}
