#include <sys/types.h>
#include <sys/fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <err.h>
#include <string.h>
#include <libgeom.h>
#include <getopt.h>
#include <md5.h>

/*
off_t f1 = 2;
off_t f2 = 2;
off_t f3 = 1;
*/

int scale = 10;

off_t f1;
off_t f2;
off_t f3;

void init_seq()
{
	f1 = 0;
	srand(1000);
}


/*
off_t next_seq()
{
	off_t f = f1 + f2/3;
	f1 = f2;
	f2 = f3;
	f3 = f;
	return f;
}
*/
off_t next_seq()
{
	f1 += rand() % (68001*scale);
	return f1;
}



int test_seq(char *devname)
{
	int fd, i;
	off_t off;
	off_t mediasize;
	size_t sectorsize;
	char *buf, ch;
	ssize_t bw;
	int count;

	fd = g_open(devname, 1);
	if (fd <= 0)
		errx(1, "Cannot open %s", devname);
	mediasize = g_mediasize(fd);
	sectorsize = g_sectorsize(fd);
	buf = malloc(sectorsize);

	printf("Writing... ");
	init_seq();
	off = count = 0;
	while (off <= mediasize) {
		off = next_seq();
		ch = (char)off;
		off = off * sectorsize;
		if (off >= mediasize - sectorsize)
			break;
		printf("%jd\t", off);
		memset(buf, ch, sectorsize);
		bw = pwrite(fd, buf, sectorsize, off);
		if (bw != sectorsize)
			errx(1, "Write failed");
		count++;
		
	}
	printf("\nWritten %d sectors\nReading... ", count);
	g_flush(fd);

	init_seq();
	off = count = 0;
	while (off <= mediasize) {
		off = next_seq();
		ch = (char)off;
		off = off * sectorsize;
		if (off >= mediasize - sectorsize) 
			break;
		printf("%jd\t", off);
		bw = pread(fd, buf, sectorsize, off);
		if (bw != sectorsize)
			errx(1, "Read failed");
		count++;
		for (i = 0; i < sectorsize; i++)
			if (buf[i] != ch)
				errx(1, "Read mismatch at offset %jd", off);
		
	}

	printf("\nRead %d sectors.\nTest ok.\n", count);
	g_close(fd);
	return 0;
}


void dump_buf(char *buf, size_t size)
{
	FILE *p = popen("/usr/bin/hd", "w");
	if (fwrite(buf, size, 1, p) != 1)
		errx(1, "Cannot write %zd bytes to hd", size);
	pclose(p);
}


/* The idea: write a block of sectors, then read subblocks and compare */
int test_overlap(char *devname) 
{
	int n_sectors = 128; /* 81920 */
	int start = 160; /* 98304 */
	int win_sectors = 32; /* 16384 */
	char *sectors, *buf;
	off_t media_size;
	int fd, sector_size;
	size_t sectorbuf_size;
	int fdr;
	int i;

	fd = g_open(devname, 1);
	if (fd <= 0)
		errx(1, "Cannot open %s", devname);
	media_size = g_mediasize(fd);
	sector_size = g_sectorsize(fd);

	sectorbuf_size = sector_size*n_sectors;
	sectors = malloc(sectorbuf_size);
	buf = malloc(sectorbuf_size);

	fdr = open("/dev/random", O_RDONLY);
	if (read(fdr, sectors, sectorbuf_size) != sectorbuf_size)
		errx(1, "read failed on /dev/random for %d", sector_size*n_sectors);
	close(fdr);
/*
	sectors_hash = malloc(n_sectors*sizeof(*sectors_hash));

	for (i = 0; i < n_sectors; i++)
		sectors_hash[i] = MD5Data(sectors+(i*sector_size), sector_size, NULL);
	*/

	/* step 1: write the whole block */
	printf("Step 1\n");
	if (pwrite(fd, sectors, sectorbuf_size, start*sector_size) != sectorbuf_size)
		err(1, "Error in pwrite for %zd at %d", sectorbuf_size, start*sector_size);

	/* step 2: read gradualy increasing subblocks */
	printf("Step 2\n");
	for (i = 1; i < n_sectors-1; i++) {
		if (pread(fd, buf, i*sector_size, start*sector_size) != i*sector_size)
			err(1, "Error in pread for %zd at %d", i*sector_size, start*sector_size);
		if (bcmp(sectors, buf, i*sector_size) != 0)
			errx(1, "COMPARE ERROR for buffer of size %zd read at %d", i*sector_size, start*sector_size);
	}

	/* step 2.5 - do some random walking on the drive */
	printf("Step 2.5\n");
	for (i = 1; i < 100; i++) {
		off_t off = random() % (media_size - sectorbuf_size);
		off = (off / sector_size) * sector_size;
		if (pwrite(fd, sectors, sectorbuf_size, off) != sectorbuf_size)
			err(1, "Error in pwrite of %zd to %jd", sectorbuf_size, off);
	}

	/* step 3: read a sliding window of 32 sectors across the block */
	printf("Step 3\n");
	for (i = 0; i < n_sectors-win_sectors; i++) {
		off_t off = (start+i)*sector_size;

		if (pread(fd, buf, win_sectors*sector_size, off) != win_sectors*sector_size)
			err(1, "Error in pread for %zd at %jd", win_sectors*sector_size, off);

		if (bcmp(sectors+i*sector_size, buf, win_sectors*sector_size) != 0) {
			printf("read:\n");
			dump_buf(buf, win_sectors*sector_size);
			printf("\n-----------------------------------------------------------\nshould be:\n");
			dump_buf(sectors+(i*win_sectors), win_sectors*sector_size);
			errx(1, "SLIDING WINDOW COMPARE ERROR for buffer of size %zd read at %jd, win ndx=%d", win_sectors*sector_size, off, i);
		}
	}


	return 0;
}


void usage(char *me) 
{
	printf("usage: %s [-o] [-s] [-x size] device\n", me);
	exit(0);
}


int main(int argc, char **argv)
{
	int ch;
	int do_seq = 0, do_overlap = 0;
	char *devname;

	if (argc < 2)
		usage(argv[0]);

	while ( (ch = getopt(argc, argv, "sohx:")) != -1) {
		switch(ch) {
		case 's': 
			do_seq = 1;
			break;
		case 'o':
			do_overlap = 1;
			break;
		case 'x':
			scale = atoi(optarg);
			break;
		default:
			usage(argv[0]);
		}

	}

	if (optind >= argc)
		errx(1, "Device name expected");

	devname = argv[optind];
	printf("Testing %s\n", devname);

	if (do_seq)
		test_seq(devname);
	if (do_overlap)
		test_overlap(devname);

	return 0;
}

