#include <sys/types.h>
#include <sys/fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <err.h>
#include <string.h>
#include <libgeom.h>
#include <getopt.h>
#include <md5.h>


#define BUF_SIZE (10*1024*1024)
#define BUF_STEP1 (32*1024)
#define BUF_STEP2 (3*1024)
#define START_OFFSET 512


char *buf;
size_t buf_size;


int main(int argc, char **argv) 
{
	int fd;
	size_t bcount = 0;
	size_t offset = START_OFFSET; /* so we'l limit ourselves to 4 GB */
	ssize_t sector_size;
	char *devname;
	char *buf2;

	if (argc < 2) {
		printf("usage: %s devname\n", argv[0]);
		exit(1);
	}
	devname = argv[1];


	buf = malloc(BUF_SIZE);
	if (buf == NULL)
		err(1, "Cannot allocate %d KB", BUF_SIZE / 1024);

	fd = open("/dev/random", O_RDONLY);
	if (read(fd, buf, BUF_SIZE) != BUF_SIZE)
		errx(1, "Cannot fetch %d KB of random data", BUF_SIZE / 1024);
	close(fd);

	fd = g_open(devname, 1);
	sector_size = g_sectorsize(fd);

	/* write BUF_SIZE random data to the device */
	while (bcount < BUF_SIZE) {
		if (pwrite(fd, buf+bcount, BUF_STEP1, offset) != BUF_STEP1)
			err(1, "Cannot write buffer of %u at %u", BUF_STEP1, offset);
		bcount += BUF_STEP1;
		offset += BUF_STEP1;
	}

	buf2 = malloc(BUF_STEP2*2);
	if (buf2 == NULL)
		err(1, "Cannot allocate %d KB", (BUF_STEP2*2)/1024);
	bcount = 0;
	offset = START_OFFSET;
	while (bcount < BUF_SIZE) {
		if (pread(fd, buf2, BUF_STEP2, offset) != BUF_STEP2)
			err(1, "Cannot read buffer of %u at %u", BUF_STEP2, offset);
		if (bcmp(buf2, buf+bcount, BUF_STEP2) != 0)
			printf("--> mismatch at %u\n", BUF_STEP2);
		bcount += BUF_STEP2*2;
		offset += BUF_STEP2*2;
	}
	close(fd);

	return 0;
}
